// Copyright 1996, 2008 by Jon Dart.  All Rights Reserved.

// Utility to read a PGN file and convert it into the opening book.
// Input is a raw PGN file.

#include "board.h"
#include "legal.h"
#include "movegen.h"
#include "options.h"
#include "movearr.h"
#include "notation.h"
#include "chessio.h"
#include "eco.h"
#include "movearr.h"
#include "strpair.h"
#include "globals.h"
#include "util.h"
#include "chessio.h"

#include <fstream>
#include <iostream>
#include <map>
#include <vector>

using namespace std;

int replace = 1;
static int sample = 0;
static int max_ply = 40;
static const int MARGIN = 75;

static void show_usage()
{
   cerr << "Usage: pgn2book -s <sample> -m <minimum ELO> -p <max ply> pgn_file1 pgn_file2 ..." << endl;
}


void write_moves(ostream &game_file, ArasanVector<StringPair> &hdrs,
const MoveArray &moves, const char *eco,
const char *result)
{
   Board board;
   int max = Util::Min(moves.num_moves(),max_ply);
   char buf[256];
   *buf = '\0';
   int last = 0;
   int num_image_size;
   for (int i = 0; i < max; i++) {
      const MoveRecord &mr = moves[i];
      num_image_size = 0;
      char img[20];
      Notation::image(board,mr.move(),img);
      const int image_size = (int)strlen(img);
      if (last + image_size + num_image_size + 1 >= MARGIN) {
         buf[last] = '\0';
         game_file << buf << endl;
         last = 0;
         *buf = '\0';
      }
      if (last != 0)
         strcat(buf," ");
      strcat(buf,img);
      last = (int)strlen(buf);
      board.doMove(mr.move());
   }
   if (last) {
      buf[last + strlen(result) + 1] = '\0';
      game_file << buf;
   }
   game_file << endl;
}


int CDECL main(int argc, char **argv)
{
   Bitboard::init();
   initOptions(argv[0]);
   Attacks::init();
   Scoring::init();
   if (!initGlobals(argv[0],0)) {
      cleanupGlobals();
      exit(-1);
   }
   atexit(cleanupGlobals);

   options.book.book_enabled = options.log_enabled = 0;
   Board board;
   ECO eco;
   int arg = 1;
   unsigned minMoves = 30;
   int minELO = 0;
   int shortg = 0;
   int games = 0;

   if (argc ==1) {
      show_usage();
      return -1;
   }
   else {
      for (;arg < argc;++arg) {
         if (argv[arg][0] == '-') {
            switch(argv[arg][1]) {
               case 'p':
                  ++arg;
                  max_ply = atoi(argv[arg]);
                  break;
               case 's':
                  ++arg;
                  sample = atoi(argv[arg]);
                  break;
               case 'm':
                  ++arg;
                  minELO = atoi(argv[arg]);
                  break;
               default:
                  cerr << "Unrecognized switch : " << argv[arg][1] << endl;
                  break;
            }
         }
         else
            break;
      }
      for (;arg < argc;arg++) {
         ifstream pgn_file( argv[arg], ios::in);
         int count = 0;
         ColorType side;
         string result, white, black;
         if (!pgn_file.good()) {
            cerr << "could not open file " << argv[arg] << endl;
            exit(-1);
         }
         else {
            ArasanVector<StringPair> hdrs;
            while (!pgn_file.eof()) {
               long first;
               MoveArray moves;
               board.reset();
               side = White;
               hdrs.removeAll();
               ChessIO::collect_headers(pgn_file,hdrs,first);
	       ++games;
               for (;;) {
                  static char buf[2048];
                  static char num[20];
                  ChessIO::Token tok = ChessIO::get_next_token(pgn_file,buf,2048);
                  if (tok == ChessIO::Eof)
                     break;
                  else if (tok == ChessIO::Number) {
                     strncpy(num,buf,20);
                  }
                  else if (tok == ChessIO::GameMove) {
                     // parse the move
                     Move m = Notation::value(board,side,buf);
                     if (IsNull(m) ||
                        !legalMove(board,StartSquare(m),
                     DestSquare(m))) {
                        // echo to both stdout and stderr
		        cerr << "Illegal move: " << buf << endl;
                        cout << "Illegal move: " << buf << endl;
                        break;
                     }
                     else {
                        BoardState bs = board.state;
                        moves.add_move(board,bs,m,"");
                        board.doMove(m);
                     }
                     side = OppositeColor(side);
                  }
                  else if (tok == ChessIO::Unknown) {
 		    cerr << "Unrecognized text: " << buf << endl;
                  }
                  else if (tok == ChessIO::Comment) {
                     // ignored for now
                  }
                  else if (tok == ChessIO::Result) {
                     result = buf;
                     break;
                  }
               }
               // done with a game
               if (moves.num_moves() < minMoves) {
		  ++shortg;
                  continue;
               }
               if (sample != 0) {
                  count += sample;
                  if (count < 100)
                     continue;
                  count = count % 100;
               }
               // output header
               string white,black,ecoC,event,site,round,date,opening,source,whiteELOStr, blackELOStr;
               ChessIO::get_header(hdrs,"ECO",ecoC);
               ChessIO::get_header(hdrs,"White",white);
               ChessIO::get_header(hdrs,"Black",black);
               ChessIO::get_header(hdrs,"Event",event);
               ChessIO::get_header(hdrs,"Site",site);
               ChessIO::get_header(hdrs,"Round",round);
               ChessIO::get_header(hdrs,"Date",date);
               ChessIO::get_header(hdrs,"Opening",opening);
               ChessIO::get_header(hdrs,"Source",source);
               if (minELO > 0) {
                  if (ChessIO::get_header(hdrs,"WhiteElo",whiteELOStr) &&
                  ChessIO::get_header(hdrs,"BlackElo",blackELOStr)) {
                     int whiteElo=0, blackElo=0;
                     if (sscanf(whiteELOStr.c_str(),"%d",&whiteElo) &&
                     sscanf(blackELOStr.c_str(),"%d",&blackElo)) {
                        if (whiteElo < minELO || blackElo < minELO)
                           continue;
                     }
                  }
                  else                            // no ELO info available
                     continue;
               }
               string header = "; ";
               if (white.length() >0) {
                  header += white;
                  header += '-';
                  header += black;
               }
               else {
                  if (opening.length() > 0) {
                     header += opening;
                     header += ' ';
                  }
                  header += source;
               }
               if (event.length() > 0 && event != "?") {
                  header += ' ';
                  if (event.length() > 50)
                     header += event.substr(0,50);
                  else
                     header += event;
               }
               if (site.length() > 0 && site != "?") {
                  header += ' ';
                  if (site.length() > 50)
                     header += site.substr(0,50);
                  else
                     header += site;
               }
               if (round.length() > 0 && round != "?") {
                  header += " (";
                  header += round;
                  header += ")";
               }
               if (date.length() > 0 && date != "?") {
                  header += ' ';
                  header += date.substr(0,4);
               }
               if (result.length() > 0 && result != "*") {
                  header += " (";
                  header += result;
                  header += ")";
               }
               if (ecoC.length() > 0) {
                  header += " [";
                  header += ecoC;
                  header += "]";
               }
               cout << "-" << endl;
               cout << header << endl;
               write_moves(cout,hdrs,moves,ecoC.c_str(),result.c_str());
            }
            pgn_file.close();
         }
      }
      cerr << "processed " << games << " games" << endl;
      cerr << "dropped " << shortg << " games (too short)" << endl;
   }

   return 0;
}
